(*
 * Lexer for SMTLIB format
 *)

{

open Smtlib_parser

(** Keyword table *)

let keyword_table = Hashtbl.create 67

let mk_keyword s tk = Hashtbl.add keyword_table s tk

let _ =
  mk_keyword "if_then_else" IF_THEN_ELSE;
  mk_keyword "and" AND;
  mk_keyword "or" OR;
  mk_keyword "xor" XOR;
  mk_keyword "not" NOT;  
  mk_keyword "iff" IFF;
  mk_keyword "implies" IMPLIES;
  mk_keyword "true" TT;
  mk_keyword "false" FF;
  mk_keyword "ite" ITE;  
  mk_keyword "distinct" DISTINCT;
  mk_keyword "let" LET;
  mk_keyword "flet" FLET;
  mk_keyword "sat" SAT;
  mk_keyword "unsat" UNSAT;
  mk_keyword "unknown" UNKNOWN;
  mk_keyword "benchmark" BENCHMARK;

  mk_keyword ":logic" LOGIC_ATTRIBUTE;
  mk_keyword ":assumption" ASSUMPTION_ATTRIBUTE;
  mk_keyword ":formula" FORMULA_ATTRIBUTE;
  mk_keyword ":status" STATUS_ATTRIBUTE;
  mk_keyword ":extrasorts" EXTRASORTS_ATTRIBUTE;
  mk_keyword ":extrapreds" EXTRAPREDS_ATTRIBUTE;
  mk_keyword ":extrafuns" EXTRAFUNS_ATTRIBUTE;
  mk_keyword ":notes" NOTES_ATTRIBUTE


let mk_attribute s = 
  try Hashtbl.find keyword_table s
  with Not_found -> OTHER_ATTRIBUTE


let mk_ident s = 
  try Hashtbl.find keyword_table s
  with Not_found -> IDENT(s)

  
open Lexing

let incr_line s =
  let pos = s.lex_curr_p in
    s.lex_curr_p <- { pos with pos_lnum = pos.pos_lnum + 1}

}


let decimal = ['0'-'9']+

let ident = ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '\'' '_' '.' '0'-'9']*

let user_value = '{' [^ '{' '}']* '}'

let string = '"' ([^ '"'] | "\\\"")* '"'


rule token = parse 
    [' ' '\t' ]+            { token lexbuf }
  | ['\n' '\r']             { incr_line lexbuf; token lexbuf }
  | ';' [^ '\n']* eof?      { token lexbuf }
  | '('                     { OPEN_PAR }
  | ')'                     { CLOSE_PAR }
  | '='                     { EQUAL }
  | "<="                    { LEQ }
  | ">="                    { GEQ }
  | "<"                     { LT }
  | ">"                     { GT }
  | '+'                     { PLUS }
  | '-'                     { MINUS }
  | '~'                     { OPPOSITE }
  | '*'                     { TIMES }
  | '/'                     { DIVIDE }
  | string                  { STRING }
  | user_value              { USER_DATA }
  | decimal                 { NUMBER(Lexing.lexeme lexbuf) }
  | '?' ident               { VAR(Lexing.lexeme lexbuf) }
  | '$' ident               { FVAR(Lexing.lexeme lexbuf) }
  | ':' ident               { mk_attribute (Lexing.lexeme lexbuf) }
  | ident                   { mk_ident (Lexing.lexeme lexbuf) }
  | _                       { LEXERROR }
