open Formula

type direction = PLUS | MINUS

let reverse_direction = function
  | MINUS -> PLUS
  | PLUS -> MINUS;;

let get_sign x =
  let r = Z.sgn x in
    if r < 0 then MINUS
    else if r > 0 then PLUS
    else assert false;;

let extract_infinity_part
    (x : int)
    (direction : direction)
    (f : linear_atomic nnf_unquantified_formula) :
  linear_atomic nnf_unquantified_formula =

  let process_expr dir l =
    if (get_sign (linear_expression_get_coefficient l x)) = dir
    then formula_true
    else formula_false in
   
  map_atomic_nnf_unquantifier_formula
    (function
       | (`Propositional _) as k -> failwith "propositional"
       | (`Greater_eq l) as w ->
	   (try process_expr direction l with Not_found -> `Atomic w))
    (function 
       | (`Propositional _) as k -> failwith "propositional"
       | (`Greater_eq l) as w ->
       ( try process_expr (reverse_direction direction) l with Not_found -> `Not (`Atomic w))) f;;

let extract_directional_expressions (var_id : int) (direction : direction)
    (f : linear_atomic nnf_unquantified_formula) : LinearExpressionSet.t =
  nnf_atomic_formula_fold
    (fun atomic already ->
      match atomic with
	| `Atomic (`Greater_eq e) when
	    (try
	       (get_sign (linear_expression_get_coefficient e var_id))
	       = direction
	     with Not_found -> false)
	    -> LinearExpressionSet.add e already
	| `Not (`Atomic (`Greater_eq e)) when
	    (try
	       (get_sign (linear_expression_get_coefficient e var_id))
	       <> direction
	     with Not_found -> false)
	    -> LinearExpressionSet.add (linear_expression_neg e) already
	| `Atomic _ | `Not (`Atomic _) -> already)
    f LinearExpressionSet.empty;;

let constraint_to_expression (var_id : int) (cte : linear_expression) =
  let coeff = linear_expression_get_coefficient cte var_id in
    assert((Z.sgn coeff) <> 0);
    if (Z.sgn coeff) < 0
    then
      {
	expr = filter_out_from_linear_expression var_id cte;
	num = Z.one;
	den = Z.neg coeff
      }
    else
      {
	expr = linear_expression_neg
	  (filter_out_from_linear_expression var_id cte);
	num = Z.one;
	den = coeff
      };;

let midpoints_of_constraints_iter (var_id : int)
    (d_plus : LinearExpressionSet.t)
    (d_minus : LinearExpressionSet.t) :
    q_linear_expression list =
  LinearExpressionSet.fold
    (fun c_minus already_minus ->
       let e_minus = constraint_to_expression var_id c_minus in
       LinearExpressionSet.fold 
	 (fun c_plus already_plus ->
	    let e_plus = constraint_to_expression var_id c_plus in
	    let e_sum = q_linear_expression_add e_minus e_plus in
	    let midpoint = q_linear_expression_reduce e_sum.expr
	      e_sum.num (Z.add e_sum.den e_sum.den) (* TODO *2 *) in
	      midpoint::already_plus)
	 d_plus already_minus) d_minus [];;

exception Unused_quantified_variable;;

let elim_existential_quantifier3 var_id f_nnf =
  let d_plus = extract_directional_expressions var_id PLUS f_nnf
  and d_minus = extract_directional_expressions var_id MINUS f_nnf in
    (*
      print_endline "Plus expressions:";
      LinearExpressionSet.iter (fun e ->
      print_linear_expression stdout var_mapper e;
      print_newline()) d_plus;
      
      print_endline "Minus expressions:";
      LinearExpressionSet.iter (fun e ->
      print_linear_expression stdout var_mapper e;
      print_newline()) d_minus;
    *)

  let i_plus = extract_infinity_part var_id PLUS f_nnf
  and i_minus = extract_infinity_part var_id MINUS f_nnf
  and f_midpoints =
    List.map (fun midpoint ->
		substitute_in_nnf_linear_formula
		  var_id midpoint f_nnf)
      (midpoints_of_constraints_iter var_id d_plus d_minus) in
  let formulas = i_plus :: i_minus :: f_midpoints in
    (*
      List.iter
      (fun f ->
      print_linear_unquantified_formula var_mapper stdout
      (f : linear_atomic nnf_unquantified_formula :> linear_unquantified_formula);
      print_newline ())
      formulas;
    *)
    mk_v_or formulas

let rec elim_existential_quantifier2 var_id = function
  | `Or l -> mk_v_or (List.map (elim_existential_quantifier2 var_id) l)
  | x -> elim_existential_quantifier3 var_id x;;

let elim_existential_quantifier1 (var_name : string)
    (f : source_unquantified_formula) : source_unquantified_formula=
  let linearization_context = empty_linearization_context () in
  let linear_f = linearize_unquantified_formula linearization_context f in
    update_trans_var_rev_table linearization_context;
  let f_nnf = to_nnf_unquantified_formula linear_f in
  let outcome =
    try
      begin
	let var_id =
	  (try Hashtbl.find linearization_context.trans_var_table var_name
	  with Not_found -> raise Unused_quantified_variable) in
	  elim_existential_quantifier2 var_id f_nnf
      end
    with Unused_quantified_variable -> f_nnf
  in unlinearize_unquantified_formula linearization_context
       (outcome : linear_atomic nnf_unquantified_formula
	:> linear_unquantified_formula);;
 
let rec elim_existential_quantifier_source vars modulo_theory f =
  match vars with
    | [] -> f
    | h::t -> elim_existential_quantifier1 h
	(elim_existential_quantifier_source t modulo_theory f);;

let rec elim_quantifiers_source (modulo_theory : source_unquantified_formula) :
    source_quantified_formula -> source_unquantified_formula =
  function
    | `And l -> `And(List.map (elim_quantifiers_source modulo_theory) l)
    | `Or l -> `Or(List.map (elim_quantifiers_source modulo_theory) l)
    | `Xor l -> `Xor(List.map (elim_quantifiers_source modulo_theory) l)
    | `Not f -> `Not (elim_quantifiers_source modulo_theory f)
    | `Ite(c, f1, f2) -> `Ite((elim_quantifiers_source modulo_theory c),
			      (elim_quantifiers_source modulo_theory f1),
			      (elim_quantifiers_source modulo_theory f2))
    | (`Atomic _) as w -> w
    | `Exists(vars, f) ->
	(elim_existential_quantifier_source vars modulo_theory
	  (elim_quantifiers_source modulo_theory f))
    | `Forall(vars, f) ->
	`Not(elim_existential_quantifier_source vars modulo_theory
	       (`Not (elim_quantifiers_source modulo_theory f)));;
