open Formula

let split_modulus = Z.of_int 1000000;;

let rec print_split_positive_integer (channel : out_channel) x =
  if (Z.compare x split_modulus) >= 0
  then let q, r = Z.fdiv_qr x split_modulus in
    output_string channel "(";
    print_split_positive_integer channel q;
    output_string channel ")*";
    output_string channel (Z.to_string split_modulus);
    output_string channel "+";
    output_string channel (Z.to_string r)
  else
    output_string channel (Z.to_string x);;

let print_split_integer (channel : out_channel) x =
  if (Z.sgn x) < 0
  then
    begin
      output_string channel "(-(";
      print_split_positive_integer channel (Z.neg x);
      output_string channel "))"
    end
  else
    begin
      output_string channel "(";
      print_split_positive_integer channel x;
      output_string channel ")"
    end
  
let rec print_lira_source_expression (channel : out_channel) : source_expression -> unit = function
  | `Sum[] -> output_string channel "0"
  | `Product[] -> output_string channel "1"
  | `Sum(l) ->
      output_string channel "(";
      print_lira_source_expression_list "+" channel l;
      output_string channel ")"
  | `Product(l) ->
      output_string channel "(";
      print_lira_source_expression_list "*" channel l;
      output_string channel ")"
  | `Quotient(x, y) ->
      output_string channel "(";
      print_lira_source_expression channel x;
      output_string channel " / ";
      print_lira_source_expression channel y;
      output_string channel ")"
  | `Neg(x) -> output_string channel "(- ";
      print_lira_source_expression channel x;
      output_string channel ")"
  | `Integer(x) -> print_split_integer channel x
  | `Variable(x) -> output_string channel x

and print_lira_source_expression_list (separator : string) (channel : out_channel) : source_expression list -> unit = function
  | [] -> ()
  | [h] -> print_lira_source_expression channel h
  | h::t ->  print_lira_source_expression channel h;
     output_string channel separator;
     print_lira_source_expression_list separator channel t;;

let print_lira_source_atomic channel : source_atomic -> unit = function
  | `Propositional x -> output_string channel x
  | `Comparison(lhs, op, rhs) ->
      print_lira_source_expression channel lhs;
      output_string channel
	(match op with
	   | `LESS -> "<"
	   | `GREATER -> ">"
	   | `LESS_EQ -> "<="
           | `GREATER_EQ -> ">="
           | `EQUAL -> "="
           | `UNEQUAL -> "!=");
      print_lira_source_expression channel rhs;;

let print_lira_quantified_formula (print_lira_atomic : out_channel -> 'atomic -> unit)
  (channel : out_channel) : (string, 'atomic) quantified_formula -> unit =

  let rec print_lira_rec = function
    | `And l -> print_lira_list_op "true" "and" l
    | `Or l -> print_lira_list_op "false" "or" l
    | `Xor l -> print_lira_list_op "false" "xor" l
    | `Not x -> output_string channel "not (";
	print_lira_rec x;
	output_string channel ")"
    | `Ite(c, f1, f2) -> output_string channel "ite (";
	print_lira_rec c;
	output_string channel ", ";
	print_lira_rec f1;
	output_string channel ", ";
	print_lira_rec f2;
	output_string channel ")"
    | `Atomic x -> print_lira_atomic channel x
    | `Forall(vars, x) ->
	print_lira_quantifier "forall" x vars
    | `Exists(vars, x) ->
	print_lira_quantifier "exists" x vars

  and print_lira_list_op neutral op l = 
    output_string channel "(";
    print_lira_list_op_rec neutral op l;
    output_string channel ")"

  and print_lira_list_op_rec neutral op = function
    | [] -> output_string channel neutral
    | [h] -> print_lira_rec h
    | h::t ->
	print_lira_rec h;
	output_string channel " ";
	output_string channel op;
	output_string channel " ";
	print_lira_list_op neutral op t

  and print_lira_quantifier quantifier formula = function
    | [] -> print_lira_rec formula
    | h::t -> output_string channel quantifier;
	output_string channel "(";
	output_string channel h;
	output_string channel " : ";
	print_lira_quantifier quantifier formula t;
	output_string channel ")"

  in print_lira_rec;;

let print_lira_source_quantified_formula :
    out_channel -> source_quantified_formula -> unit =
  print_lira_quantified_formula print_lira_source_atomic;;

Multiple_choices.add_choice output_syntax "lira" 10 print_lira_source_quantified_formula "LIRA syntax";;
