open Formula

let minimize_generalized1 = ref false
and minimize_generalized2 = ref false
and debug = ref false
and show_conjunction_sizes = ref false
and full_simplify = ref false
and block_projected_model = ref true
and add_blocking_to_g = ref false;;

Mjollnir_options.add_bool_option  "minimize-generalized1" minimize_generalized1
  "Minimize constraints at step 1";
Mjollnir_options.add_bool_option  "minimize-generalized2" minimize_generalized2
  "Minimize constraints at step 2";
Mjollnir_options.add_bool_option  "qelim-trace" debug
  "Trace quantifier elimination";
Mjollnir_options.add_bool_option  "show-conjunction-sizes"
  show_conjunction_sizes
  "Show conjunction sizes";
Mjollnir_options.add_bool_option  "full-simplify" full_simplify
  "Full simplification";;
Mjollnir_options.add_bool_option  "block-projected-model" block_projected_model
  "Add projected model (instead of model) as blocking clause when eliminating quantifiers";;
Mjollnir_options.add_bool_option  "add-blocking-to-g" add_blocking_to_g
  "Add blocking clause to negative context (g)";;


let mk_xor2 context a b =
  Yices.mk_or context
    [Yices.mk_and context [Yices.mk_not context a; b];
     Yices.mk_and context [Yices.mk_not context b; a]];;

let rec mk_xor context = function
  | [] -> Yices.mk_true context
  | [x] -> x
  | h::t -> mk_xor2 context h (mk_xor context t);;

let rec translate_unquantified_formula
  (translate_atomic : 'atomic -> Yices.expr)
  (context : Yices.context) :
  'atomic unquantified_formula -> Yices.expr = function
    | `And [] -> Yices.mk_true context
    | `And(l) -> Yices.mk_and context
	(List.map
	   (translate_unquantified_formula translate_atomic context) l)

    | `Or [] -> Yices.mk_false context
    | `Or(l) -> Yices.mk_or context
	(List.map
	   (translate_unquantified_formula translate_atomic context) l)

    | `Xor [] -> Yices.mk_false context
    | `Xor(l) -> mk_xor context
	(List.map
	   (translate_unquantified_formula translate_atomic context) l)

    | `Ite(c, f1, f2) ->
	Yices.mk_ite context
	  (translate_unquantified_formula translate_atomic context c)
	  (translate_unquantified_formula translate_atomic context f1)
	  (translate_unquantified_formula translate_atomic context f2)
    | `Not(x) -> Yices.mk_not context
	(translate_unquantified_formula translate_atomic context x)
    | `Atomic x -> translate_atomic x;;

type source_translate_context = {
  source_translate_propositional_map : (string, Yices.var_decl) Hashtbl.t;
  source_translate_real_map : (string, Yices.var_decl) Hashtbl.t;
  source_translate_context : Yices.context };;

let mk_source_translate_context (context : Yices.context) :
    source_translate_context = {
  source_translate_context = context;
  source_translate_propositional_map = Hashtbl.create 100;
  source_translate_real_map = Hashtbl.create 100 };;

let translate_source_real_var (context: source_translate_context)
  (x : string) : Yices.var_decl =
  try Hashtbl.find context.source_translate_real_map x
  with Not_found ->
    let decl = Yices.mk_real_var_decl context.source_translate_context x in
      Hashtbl.add context.source_translate_real_map x decl;
      decl;;

let rec translate_source_expression (context : source_translate_context) : source_expression -> Yices.expr = function
  | `Sum [] -> Yices.mk_num context.source_translate_context 0
  | `Sum l -> Yices.mk_sum context.source_translate_context
       (List.map (translate_source_expression context) l)
  | `Product [] -> Yices.mk_num context.source_translate_context 1
  | `Product l -> Yices.mk_mul context.source_translate_context
       (List.map (translate_source_expression context) l)
  | `Neg x -> Yices.mk_sub context.source_translate_context
       [Yices.mk_num context.source_translate_context 0;
        translate_source_expression context x]
  | `Quotient(x, y) -> failwith "quotient"
  | `Integer x -> Yices.mk_num_from_mpz context.source_translate_context x
  | `Variable x -> Yices.mk_var_from_decl context.source_translate_context
      (translate_source_real_var context x);;

let translate_source_bool_var (context: source_translate_context)
  (x : string) : Yices.var_decl =
  try Hashtbl.find context.source_translate_propositional_map x
  with Not_found ->
    let decl = Yices.mk_bool_var_decl context.source_translate_context x in
      Hashtbl.add context.source_translate_propositional_map x decl;
      decl;;

let translate_source_atomic (context : source_translate_context) : source_atomic -> Yices.expr = function
  | `Propositional x -> Yices.mk_var_from_decl context.source_translate_context
      (translate_source_bool_var context x)
  | `Comparison(lhs, op, rhs) ->
      (match op with
	 | `LESS -> Yices.mk_lt
	 | `LESS_EQ -> Yices.mk_le
	 | `GREATER -> Yices.mk_gt
	 | `GREATER_EQ -> Yices.mk_ge
	 | `EQUAL -> Yices.mk_eq
	 | `UNEQUAL -> Yices.mk_diseq) context.source_translate_context
	(translate_source_expression context lhs)
	(translate_source_expression context rhs);;

let translate_source_formula (context : source_translate_context) :
  source_unquantified_formula -> Yices.expr =
  translate_unquantified_formula
    (translate_source_atomic context) context.source_translate_context;;

let eval_source_formula_in_yices_model (context : source_translate_context)
  (model : Yices.model) (formula : source_unquantified_formula) =
  eval_source_formula
    (fun real_var -> Yices.get_mpq_value model
       (Hashtbl.find context.source_translate_real_map real_var))
    (fun prop_var -> Yices.bool_of_lbool (Yices.get_value model
	 (Hashtbl.find context.source_translate_propositional_map prop_var)))
    formula;;

type linear_translate_context = {
  linear_translate_atomic_map :
    (linear_atomic, Yices.expr) Hashtbl.t;
  linear_translate_real_map : Yices.var_decl array;
  linear_translate_context : Yices.context };;

let translate_linear_expression (context : linear_translate_context)
  (l : linear_expression) : Yices.expr =
  Yices.mk_sum context.linear_translate_context
    (List.map
       (fun (id, coeff) ->
	  if id < 0
	  then Yices.mk_num_from_mpz context.linear_translate_context coeff
	  else
	    let var =
	      Yices.mk_var_from_decl context.linear_translate_context
		context.linear_translate_real_map.(id) in
	      if Z.equal coeff Z.one
	      then var
	      else Yices.mk_mul context.linear_translate_context
		[Yices.mk_num_from_mpz context.linear_translate_context coeff;
		 var ]) l);;

let translate_linear_atomic_uncached (context : linear_translate_context) :
    linear_atomic -> Yices.expr = function
      | `Propositional x ->
	  Yices.mk_var_from_decl context.linear_translate_context
	    (Yices.mk_bool_var_decl context.linear_translate_context x)
      | `Greater_eq l ->
	  (match l with
	     | [] -> Yices.mk_true context.linear_translate_context
	     | [-1, cst] ->
		 if (Z.sgn cst) >= 0
		 then Yices.mk_true context.linear_translate_context
		 else Yices.mk_false context.linear_translate_context
	     | _ ->
		 Yices.mk_ge context.linear_translate_context
		   (translate_linear_expression context l)
		   (Yices.mk_num context.linear_translate_context 0));;

let translate_linear_atomic (context : linear_translate_context)
    (a : linear_atomic) =
  try Hashtbl.find context.linear_translate_atomic_map a
  with Not_found ->
    let x = translate_linear_atomic_uncached context a in
    Hashtbl.add context.linear_translate_atomic_map a x; x;;

let translate_linear_formula (context : linear_translate_context) :
  linear_unquantified_formula -> Yices.expr =
  translate_unquantified_formula
    (translate_linear_atomic context) context.linear_translate_context;;

let mk_linear_translate_context (context : Yices.context)
    (lzc : linearization_context) : linear_translate_context =
  {
    linear_translate_atomic_map = Hashtbl.create 100;
    linear_translate_real_map =
      Array.map (fun name -> Yices.mk_real_var_decl context name)
	lzc.trans_var_rev_table;
    linear_translate_context = context
  }

let eval_linear_formula_in_yices_model (context : linear_translate_context)
  (model : Yices.model) (formula : linear_unquantified_formula) =
  eval_linear_formula
    (fun real_var -> Yices.get_mpq_value model
       context.linear_translate_real_map.(real_var))
    (fun prop_var -> failwith "eval_linear_formula_in_yices_model")
    formula;;

type smt_model = {
  real_model : int -> Q.t;
  bool_model : string -> bool
};;

type generalized_model = {
  generalized_real_model : [`Greater_eq of linear_expression] list;
  generalized_bool_model : string -> bool
};;

type 'model smt_answer =
  | UNSATISFIABLE
  | SATISFIABLE of 'model;;

let linear_solve (context : linear_translate_context) : smt_model smt_answer =
  match Yices.check context.linear_translate_context with
    | Yices.L_TRUE ->
	let yices_model = Yices.get_model context.linear_translate_context in
	  SATISFIABLE
	  {
	    real_model =
	      (let m = Array.map
		(fun decl -> Yices.get_mpq_value yices_model decl)
		context.linear_translate_real_map in
		fun i -> m.(i));
	    bool_model =
	      fun x -> failwith "bool_model"
	  }
		 
    | Yices.L_FALSE -> UNSATISFIABLE
    | Yices.L_UNKNOWN -> failwith "undecidable model???";;

let eval_linear_formula_in_smt_model
  (model : smt_model)
  (formula : linear_unquantified_formula) : bool =
  eval_linear_formula model.real_model model.bool_model formula;;

type dual_linear_translate_context = {
  context_formula : linear_unquantified_formula;
  positive_context : linear_translate_context;
  negative_context : linear_translate_context };;

let mk_dual_context (lc : linearization_context)
    (modulo_theory : linear_unquantified_formula)
    (linearized_formula : linear_unquantified_formula) =
  let dc = {
    context_formula = linearized_formula;
    positive_context = mk_linear_translate_context (Yices.mk_context ()) lc;
    negative_context = mk_linear_translate_context (Yices.mk_context ()) lc} in

  Yices.assert_expr dc.positive_context.linear_translate_context
    (translate_linear_formula
       dc.positive_context modulo_theory);
  Yices.assert_expr dc.negative_context.linear_translate_context
    (translate_linear_formula
       dc.negative_context modulo_theory);

  Yices.assert_expr dc.positive_context.linear_translate_context
    (translate_linear_formula
       dc.positive_context linearized_formula);
  Yices.assert_expr dc.negative_context.linear_translate_context
    (translate_linear_formula
       dc.negative_context (`Not linearized_formula));
  dc;;

let generalize_model1 (dc : dual_linear_translate_context)
    (model : smt_model) =
  let atomic_formulas =
    collect_linear_atomic_formulas_in_unquantified dc.context_formula in
       List.map
	 (fun expression ->
	    (if (Q.sgn (eval_linear_expression model.real_model
			  expression)) >= 0
	     then expression, `GREATER_EQ
	     else expression, `LESS))
	 (LinearExpressionSet.elements atomic_formulas);;

let print_model channel linearization_context (model : smt_model) =
  let first = ref true in
  Array.iteri (fun i name ->
		(if not !first then output_string channel ", ");
		Printf.fprintf channel "%s = %s" name
		  (Q.to_string (model.real_model i)); 
		first := false
	     ) linearization_context.trans_var_rev_table;;

let print_model2 channel linearization_context ctx =
  let yices_model = Yices.get_model ctx.linear_translate_context in
  Array.iteri (fun i decl ->
		Printf.printf "%s = %s\n"
		  linearization_context.trans_var_rev_table.(i)
		  (Q.to_string (Yices.get_mpq_value yices_model decl))
	     ) ctx.linear_translate_real_map;;

(* TODO: Warn SRI, apparently this procedure does not work if context
  is already contradictory *)
let test_coherence linearization_context ctx (f : Yices.expr) =
  assert ((Yices.check ctx.linear_translate_context) = Yices.L_TRUE);
  Yices.push ctx.linear_translate_context;
  Yices.assert_expr ctx.linear_translate_context f;
  let r =
    (match Yices.check ctx.linear_translate_context with
    | Yices.L_TRUE -> true
    | Yices.L_FALSE -> false
    | Yices.L_UNKNOWN -> failwith "unknown in test_coherence") in
    Yices.pop ctx.linear_translate_context;
    r;;

let conjunct_to_atomic 
    ((f : linear_expression),
     (op : convex_comparison_op))  =
  match op with
    | `GREATER_EQ -> `Atomic(`Greater_eq f)
    | `LESS -> `Not(`Atomic(`Greater_eq f))
    | `LESS_EQ -> `Atomic(`Greater_eq (linear_expression_neg f))
    | `GREATER -> `Not(`Atomic(`Greater_eq (linear_expression_neg f)))
    | `EQUAL -> `And [`Atomic(`Greater_eq f);
		      `Atomic(`Greater_eq (linear_expression_neg f))];;

let rec merge_and (already : 'atomic literal list) = function
  | [] -> already
  | (`And [])::t -> merge_and already t
  | (`And ((h1 : 'atomic literal)::t1))::t ->
      merge_and (h1::already) ((`And t1)::t)
  | ((`Atomic _ | `Not (`Atomic _)) as w)::t -> merge_and (w::already) t;;

let conjuncts_to_conjunction
    (conjuncts : (linear_expression * convex_comparison_op) list) :
    linear_atomic conjunction =
  `And(List.sort Pervasives.compare
	 (merge_and [] (List.map conjunct_to_atomic conjuncts)));;

let conjuncts_to_formula
    (conjuncts : (linear_expression * convex_comparison_op) list) :
    'atomic unquantified_formula =
  ((conjuncts_to_conjunction conjuncts) : 'atomic conjunction
   :> 'atomic unquantified_formula);;

(* Try generalization sequentially: left from right *)
let rec generalize_model2_process linearization_context
    (negative_c : linear_translate_context)
    already_decided to_be_decided =
  Minimal.generalize (fun x -> not (test_coherence linearization_context negative_c (translate_linear_formula negative_c (conjuncts_to_formula
										    x ))))
    already_decided to_be_decided;;
	  
let generalize_model2 linearization_context
    (negative_c : linear_translate_context)
    (in_model :
       (linear_expression *
	  [`GREATER_EQ | `GREATER | `LESS_EQ | `LESS | `EQUAL]) list) :
  (linear_expression *
     [`GREATER_EQ | `GREATER | `LESS_EQ | `LESS | `EQUAL]) list =
  let out_model = generalize_model2_process linearization_context
    negative_c [] in_model in
    (if !show_conjunction_sizes
     then
       (Printf.printf "in: %d conjuncts, out: %d conjuncts\n"
	 (List.length in_model) (List.length out_model);
	flush stdout));
    out_model;;

let conjuncts_inject l=
  (l : (Formula.linear_expression * [ `EQUAL | `GREATER | `GREATER_EQ ]) list
   :> (Formula.linear_expression * Formula.convex_comparison_op) list);;

let canonify_inequalities = List.map
  (function
     | ((_, `EQUAL) | (_, `GREATER) | (_, `GREATER_EQ)) as ineq -> ineq
     | (expr, `LESS) -> (linear_expression_neg expr), `GREATER
     | (expr, `LESS_EQ) -> (linear_expression_neg expr), `GREATER_EQ
  );;

let generalize_model linearization_context
    (dc : dual_linear_translate_context) (model : smt_model):
    (linear_expression *
     [`GREATER_EQ | `GREATER | `EQUAL]) list =

  (if !debug then
     begin
       print_string "Non generalized: ";
       print_model stdout linearization_context model;
       print_newline()
     end);

  let generalized1 = generalize_model1 dc model in
    (if !debug then
       begin
	 print_string "Generalized1:  ";
	 Formula_math.print_processed_formula stdout linearization_context (conjuncts_to_formula (generalized1 :> (linear_expression * convex_comparison_op) list));
	 print_newline();
       end);
    assert (not (test_coherence linearization_context dc.negative_context
		   (translate_linear_formula dc.negative_context
		      (conjuncts_to_formula generalized1))));
    let generalized1m =
      if !minimize_generalized1
      then conjuncts_inject (Projection.minimize_constraints generalized1)
      else generalized1 in

      assert(not(test_coherence linearization_context dc.negative_context
		   (translate_linear_formula dc.negative_context
		      (conjuncts_to_formula generalized1m))));

      let generalized2 = generalize_model2 linearization_context
	dc.negative_context generalized1m in
	assert(not(test_coherence linearization_context dc.negative_context
		     (translate_linear_formula dc.negative_context
			(conjuncts_to_formula generalized2))));

	let generalized2m =
	  if !minimize_generalized2
	  then Projection.minimize_constraints generalized2
	  else canonify_inequalities generalized2 in
	  assert(not(test_coherence linearization_context dc.negative_context
		       (translate_linear_formula dc.negative_context
			  (conjuncts_to_formula (conjuncts_inject generalized2m)))));
	  generalized2m;;

let get_a_generalized_model linearization_context
    (dc : dual_linear_translate_context) =
  match linear_solve dc.positive_context with
    | UNSATISFIABLE -> UNSATISFIABLE
    | SATISFIABLE model -> SATISFIABLE (generalize_model linearization_context dc model);;

let add_blocking_clause dc formula =
  Yices.assert_expr dc.positive_context.linear_translate_context
    (translate_linear_formula dc.positive_context (`Not formula));
  (if !add_blocking_to_g
  then
    Yices.assert_expr dc.negative_context.linear_translate_context
      (translate_linear_formula dc.negative_context (`Not formula)));;

let is_always_true dc =
  match Yices.check dc.negative_context.linear_translate_context with
    | Yices.L_TRUE -> false
    | Yices.L_FALSE -> true
    | Yices.L_UNKNOWN -> failwith "undecidable in is_always_true";;

let rec get_all_models linearization_context dc already =
  if is_always_true dc
  then [[]]
  else 
    match get_a_generalized_model linearization_context dc with
      | UNSATISFIABLE -> already
      | SATISFIABLE model ->
	  add_blocking_clause dc
	    (conjuncts_to_formula (conjuncts_inject model));
	  get_all_models linearization_context dc (model::already);;

let is_satisfiable linearization_context formula =
  let lc =
    mk_linear_translate_context (Yices.mk_context ()) linearization_context in
    Yices.assert_expr lc.linear_translate_context
      (translate_linear_formula lc formula);
    match Yices.check lc.linear_translate_context with
      | Yices.L_TRUE -> true
      | Yices.L_FALSE -> false
      | Yices.L_UNKNOWN -> failwith "is_satisfiable";;
  

let to_dnf_unchecked linearization_context modulo_theory formula :
    linear_atomic dnf_formula =
  let dc = mk_dual_context linearization_context modulo_theory formula in
    `Or (List.sort Pervasives.compare
	   (List.map (fun x -> conjuncts_to_conjunction (conjuncts_inject x))
	   (get_all_models linearization_context dc [])));;

let to_dnf linearization_context modulo_theory formula =
  let dnf = to_dnf_unchecked linearization_context modulo_theory formula in
    assert(not (is_satisfiable linearization_context
		  (`And [(`Not formula);
			  (dnf : linear_atomic dnf_formula
			   :> linear_unquantified_formula)])));
    assert(not (is_satisfiable linearization_context
		  (`And [formula; (`Not
                          (dnf : linear_atomic dnf_formula
                           :> linear_unquantified_formula))])));
    dnf;;

let rec get_all_projected_models linearization_context dc dimension already =
  if is_always_true dc
  then [[]]
  else 
    match get_a_generalized_model linearization_context dc with
      | UNSATISFIABLE -> already
      | SATISFIABLE model ->
	  let model' =
	    Projection.project_minimize_constraints
	      (conjuncts_inject model) dimension in
	  add_blocking_clause dc
	    (conjuncts_to_formula (conjuncts_inject
	      (if !block_projected_model
               then model'
               else model)));
	  get_all_projected_models linearization_context dc
	    dimension (model'::already);;

let elim_existential_quantifier_linear linearization_context modulo_theory dimension formula : linear_atomic dnf_formula =
  assert(dimension >= 0);
  assert(dimension <= linearization_context.trans_nr_vars);
  let dc = mk_dual_context linearization_context modulo_theory formula in
    `Or (List.map (fun x -> conjuncts_to_conjunction (conjuncts_inject x))
	   (get_all_projected_models linearization_context dc dimension []));;

let stringset_of_list (l : string list) : StringSet.t =
  List.fold_right StringSet.add l StringSet.empty;;

let prepare_projection_linearization_context (elim_variables : string list)
    (source_formula : source_unquantified_formula) :
    int*linearization_context =
  let variables = collect_source_variables_in_unquantified source_formula in
  let quantified_variables = StringSet.inter variables
    (stringset_of_list elim_variables) in
  let free_variables = StringSet.diff variables quantified_variables in
  let free_variables_l = StringSet.elements free_variables in
  let trans_var_rev_table = Array.of_list (List.append
					     free_variables_l (StringSet.elements quantified_variables)) in
  let trans_var_table = Hashtbl.create 100 in
    Array.iteri (fun i name -> Hashtbl.add trans_var_table name i)
      trans_var_rev_table;
    (List.length free_variables_l),
    {
      trans_nr_vars = Array.length trans_var_rev_table;
      trans_var_table = trans_var_table;
      trans_var_rev_table = trans_var_rev_table };;
  
(* TODO bug if modulo_theory has quantified variables *)
let elim_existential_quantifier_source (elim_variables : string list)
    (modulo_theory : source_unquantified_formula)
    (source_formula : source_unquantified_formula) =
  let nr_free_variables, linearization_context =
    prepare_projection_linearization_context
    elim_variables (`And [modulo_theory; source_formula]) in
  let linear_formula = linearize_unquantified_formula
    linearization_context source_formula
  and linear_modulo_theory = linearize_unquantified_formula
    linearization_context modulo_theory in
    unlinearize_dnf_formula linearization_context
      (elim_existential_quantifier_linear linearization_context
	 linear_modulo_theory
	 nr_free_variables linear_formula);;

let rec elim_quantifiers_source (modulo_theory : source_unquantified_formula) :
    source_quantified_formula -> source_unquantified_formula =
  function
    | `And l -> `And(List.map (elim_quantifiers_source modulo_theory) l)
    | `Or l -> `Or(List.map (elim_quantifiers_source modulo_theory) l)
    | `Xor l -> `Xor(List.map (elim_quantifiers_source modulo_theory) l)
    | `Not f -> `Not (elim_quantifiers_source modulo_theory f)
    | `Ite(c, f1, f2) -> `Ite((elim_quantifiers_source modulo_theory c),
			      (elim_quantifiers_source modulo_theory f1),
			      (elim_quantifiers_source modulo_theory f2))
    | (`Atomic _) as w -> w
    | `Exists(vars, f) ->
	(elim_existential_quantifier_source vars modulo_theory
	  (elim_quantifiers_source modulo_theory f) :
	  source_dnf_formula :> source_unquantified_formula)
    | `Forall(vars, f) ->
	`Not(elim_existential_quantifier_source vars modulo_theory
	       (`Not (elim_quantifiers_source modulo_theory f)) :
	    source_dnf_formula :> source_unquantified_formula);;


let reduction_step (modulo_theory : source_unquantified_formula)
    (f : source_quantified_formula) : source_dnf_formula =
  elim_existential_quantifier_source [] modulo_theory
    (elim_quantifiers_source modulo_theory (`Forall([], f)));;

let rec reduce_source_dnf (modulo_theory : source_unquantified_formula)
    (f : source_dnf_formula) =
  let f2 = reduction_step modulo_theory
    (f : source_dnf_formula :> source_quantified_formula) in
    if (quantified_formula_size
	  (f2 : source_dnf_formula :> source_quantified_formula)) <
      (quantified_formula_size
	 (f : source_dnf_formula :> source_quantified_formula))
    then reduce_source_dnf modulo_theory f2
    else f;;

let reduce_dnf_source_formula (modulo_theory : source_unquantified_formula)
    (f : source_quantified_formula) =
  let f1 = elim_existential_quantifier_source [] modulo_theory
    (elim_quantifiers_source modulo_theory f) in
    reduce_source_dnf modulo_theory f1;;

let rec reduce (modulo_theory : source_unquantified_formula)
  (s : source_quantified_formula) : source_unquantified_formula =
  if !full_simplify
  then ((reduce_dnf_source_formula modulo_theory s) :
	  Formula.source_dnf_formula :> Formula.source_unquantified_formula)
  else elim_quantifiers_source modulo_theory s;;
