open Formula

type direction = Ferrante_rackoff.direction
let get_sign = Ferrante_rackoff.get_sign
let constraint_to_expression = Ferrante_rackoff.constraint_to_expression
let extract_infinity_part = Ferrante_rackoff.extract_infinity_part

let extract_wide_directional_expressions (var_id : int) (direction : direction)
  (f : linear_atomic nnf_unquantified_formula) : LinearExpressionSet.t =
  nnf_atomic_formula_fold
    (fun atomic already ->
      match atomic with
	| `Atomic (`Greater_eq e) when
	    (try get_sign (linear_expression_get_coefficient e var_id)=direction
	    with Not_found -> false)
	    -> LinearExpressionSet.add e already
	| `Not (`Atomic _) | `Atomic _ -> already)
    f LinearExpressionSet.empty;;

let extract_strict_directional_expressions (var_id : int) (direction : direction)
    (f : linear_atomic nnf_unquantified_formula) : LinearExpressionSet.t =
  nnf_atomic_formula_fold
    (fun atomic already ->
      match atomic with
	| `Not (`Atomic (`Greater_eq e)) when
	    (try
	       (get_sign (linear_expression_get_coefficient e var_id))
	       <> direction
	     with Not_found -> false)
	    -> LinearExpressionSet.add e already
	| `Atomic _ | `Not (`Atomic _) -> already)
    f LinearExpressionSet.empty;;

let epsilon_substitution
    (x : int)
    (replacement : q_linear_expression)
    (direction : direction)
    (f : linear_atomic nnf_unquantified_formula) :
  linear_atomic nnf_unquantified_formula =

  let process_expr invert e =
    if (get_sign (linear_expression_get_coefficient e x)) = direction
    then
      (let e' = substitute_in_q_linear_expression x
	  replacement { expr=e; num=Z.one; den=Z.one } in
	  assert((Z.sgn e'.num) > 0);
	  assert((Z.sgn e'.den) > 0);
	  (if invert then gen_less else gen_greater_eq) e'.expr)
    else
      (let e' = substitute_in_q_linear_expression x
	  replacement { expr=e; num=Z.one; den=Z.one } in
	  assert((Z.sgn e'.num) > 0);
	  assert((Z.sgn e'.den) > 0);
	  (* print_string "x = ";
	  Formula_math.print_q_linear_expression stdout (fun x -> Printf.sprintf "v%d" x) replacement;
          print_newline ();
	  print_string "e = ";
	  Formula_math.print_linear_expression stdout (fun x -> Printf.sprintf "v%d" x) e;
          print_newline (); *)

	  (if invert then gen_greater_eq else gen_less)
	    (linear_expression_neg e'.expr)) in
   
  map_atomic_nnf_unquantifier_formula
    (function
       | `Propositional _ -> failwith "propositional"
       | (`Greater_eq l) as w ->
	   (try process_expr false l with Not_found -> `Atomic w))
    (function 
       | `Propositional _ -> failwith "propositional"
       | (`Greater_eq l) as w ->
       ( try process_expr true l with Not_found -> `Not (`Atomic w))) f;;
    
let elim_existential_quantifier3 direction var_id f_nnf =
  let i_plus = extract_infinity_part var_id Ferrante_rackoff.PLUS f_nnf
  and i_minus = extract_infinity_part var_id Ferrante_rackoff.MINUS f_nnf
  and d_wide_edges =  extract_wide_directional_expressions var_id direction f_nnf
  and d_strict_edges = extract_strict_directional_expressions var_id direction f_nnf in
  let f_wide_edges = List.map
    (fun d_wide_edge ->
       substitute_in_nnf_linear_formula var_id
	 (constraint_to_expression var_id d_wide_edge) f_nnf)
    (LinearExpressionSet.elements d_wide_edges )
  and f_strict_edges = List.map
    (fun d_strict_edge ->
       epsilon_substitution var_id
	 (constraint_to_expression var_id d_strict_edge)
	 direction f_nnf)
    (LinearExpressionSet.elements d_strict_edges )
  in
  mk_v_or ((mk_v_or (i_plus :: i_minus :: f_strict_edges))::f_wide_edges);;

let rec elim_existential_quantifier2 direction var_id = function
  | `Or l -> mk_v_or (List.map (elim_existential_quantifier2 direction var_id) l)
  | x -> elim_existential_quantifier3 direction var_id x;;

let elim_existential_quantifier1 direction (var_name : string)
    (f : source_unquantified_formula) : source_unquantified_formula=
  let linearization_context = empty_linearization_context () in
  let linear_f = linearize_unquantified_formula linearization_context f in
    update_trans_var_rev_table linearization_context;
  let f_nnf = to_nnf_unquantified_formula linear_f in
  let outcome =
    try
      begin
	let var_id =
	  (try Hashtbl.find linearization_context.trans_var_table var_name
	  with Not_found -> raise Ferrante_rackoff.Unused_quantified_variable) in
	  elim_existential_quantifier2 direction var_id f_nnf
      end
    with Ferrante_rackoff.Unused_quantified_variable -> f_nnf
  in unlinearize_unquantified_formula linearization_context
       (outcome : linear_atomic nnf_unquantified_formula
	:> linear_unquantified_formula);;
 
let rec elim_existential_quantifier_source direction vars modulo_theory f =
  match vars with
    | [] -> f
    | h::t -> elim_existential_quantifier1 direction h
	(elim_existential_quantifier_source direction t modulo_theory f);;

let rec elim_quantifiers_source (modulo_theory : source_unquantified_formula) :
    source_quantified_formula -> source_unquantified_formula =
  function
    | `And l -> `And(List.map (elim_quantifiers_source modulo_theory) l)
    | `Or l -> `Or(List.map (elim_quantifiers_source modulo_theory) l)
    | `Xor l -> `Xor(List.map (elim_quantifiers_source modulo_theory) l)
    | `Not f -> `Not (elim_quantifiers_source modulo_theory f)
    | `Ite(c, f1, f2) -> `Ite((elim_quantifiers_source modulo_theory c),
			      (elim_quantifiers_source modulo_theory f1),
			      (elim_quantifiers_source modulo_theory f2))
    | (`Atomic _) as w -> w
    | `Exists(vars, f) ->
	(elim_existential_quantifier_source Ferrante_rackoff.PLUS vars modulo_theory
	  (elim_quantifiers_source modulo_theory f))
    | `Forall(vars, f) ->
	`Not(elim_existential_quantifier_source Ferrante_rackoff.PLUS vars modulo_theory
	       (`Not (elim_quantifiers_source modulo_theory f)));;
