open Formula;;

let rec find_assignment (computed_variable : int) = function
  | [] -> failwith "cannot find assignment"
  | (linear_expr, `GREATER_EQ)::rest ->
      (try
	let coeff = linear_expression_get_coefficient linear_expr computed_variable in
	  gen_q_linear_expression
	    (filter_out_from_linear_expression computed_variable linear_expr)
	    (Z.neg coeff)
       with Not_found -> find_assignment computed_variable rest)
  | _::rest -> find_assignment computed_variable rest;;

let add_counter expr already =
  LinearExpressionMap.add expr
    (succ (try LinearExpressionMap.find expr already
	   with Not_found -> 0)) already;;

let collect_constraint_expressions1 already (expr, op) =
  match op with
    | `GREATER_EQ -> add_counter expr already
    | `GREATER -> add_counter (linear_expression_neg expr) already
    | `EQUAL -> add_counter (linear_expression_neg expr)
	(add_counter expr already);;

let collect_constraint_expressions2 already : Projection.minimize_out ->
  int LinearExpressionMap.t = List.fold_left collect_constraint_expressions1 already;;

let map_to_list (v : 'a LinearExpressionMap.t) : (linear_expression * 'a) list=
  LinearExpressionMap.fold (fun key data already ->
			      (key, data)::already) v [];;

exception Empty_list

let rec max_of_list : ('a*int) list -> 'a*int = function
  | [] -> raise Empty_list
  | [w] -> w
  | ((hv, hi) as hw)::t ->
      let (tv, ti) as tw = max_of_list t in
      if hi > ti
      then hw
      else tw;;

let collect_constraint_expressions3 f =
  map_to_list (List.fold_left collect_constraint_expressions2 LinearExpressionMap.empty f);;

type decision_tree =
  | IfThenElse of linear_expression * decision_tree * decision_tree
  | Defined of q_linear_expression
  | Undefined;;

exception Found_definition of q_linear_expression;;

let conjuncts_to_constraints conjuncts =
  List.map (function
     |`Atomic (`Propositional _)
     | `Not (`Atomic (`Propositional _)) ->
	 failwith "propositional"
     | `Atomic (`Greater_eq x) -> (x, `GREATER_EQ) 
     | `Not (`Atomic (`Greater_eq x)) -> (x, `LESS))
    conjuncts;;

let build_if_then_else splitter a b = IfThenElse(splitter, a, b)

let rec build_decision_tree linearization_context (computed_variable : int)
    (linear_modulo_theory : linear_unquantified_formula)
    (linear_formula : linear_unquantified_formula) =
  let `Or disjuncts = Formula_yices.elim_existential_quantifier_linear
	linearization_context linear_modulo_theory (computed_variable + 1)
	linear_formula in

    match disjuncts with
      | [] -> Undefined
      | _ ->
	  try
	    begin
	      let definition_blocks =
		List.map (fun (`And conjuncts) ->
			    let constraints = conjuncts_to_constraints conjuncts in
			      match
				Projection.project_minimize_constraints constraints
				  computed_variable
			      with
				| [] ->
				    raise (Found_definition(
				      find_assignment computed_variable constraints))
				| projected_constraints -> projected_constraints
			 ) disjuncts in
	      let splitter, _ =
		try max_of_list
		  (collect_constraint_expressions3 definition_blocks)
	      with Empty_list -> failwith "incomprehensible" in
	      
	      let create_branch cond =
		build_decision_tree linearization_context computed_variable
		  (mk_v_and [cond; linear_modulo_theory])
		  (mk_v_and [cond; linear_formula]) in
		
	      build_if_then_else splitter
		(create_branch (`Atomic (`Greater_eq splitter)))
		(create_branch (`Not (`Atomic (`Greater_eq splitter))))
	    end
	  with Found_definition x -> Defined x;;

let rec print_decision_tree (channel : out_channel) 
    (var_mapper : int->string) indent = function
  | Defined x -> output_string channel indent;
      Formula_math.print_q_linear_expression channel var_mapper x
  | Undefined -> output_string channel indent;
      output_string channel "Failure"
  | IfThenElse(cond, a, b) ->
      output_string channel indent;
      output_string channel "If[";
      Formula_math.print_linear_expression channel var_mapper cond;
      output_string channel " >= 0,\n";
      print_decision_tree channel var_mapper (indent ^ "  ") a;
      output_string channel ",\n";
      print_decision_tree channel var_mapper (indent ^ "  ") b;
      output_string channel "]";;
