open Apron;;

let coeff_of_z (coeff : Z.t) : Coeff.t =
  Coeff.Scalar (Scalar.Mpqf (Q.from_z coeff));;

let apron_linear_exprof_linear_expr_internal (constant_coeff : Coeff.t option)
    (l : Formula.linear_expression) =
  let expr = Linexpr0.make None in
    Linexpr0.set_list expr (List.map (fun (id, coeff) ->
					(coeff_of_z coeff), id) l)
      constant_coeff;
  expr;;

let apron_linexpr0_of_linear_expr : Formula.linear_expression -> Linexpr0.t =
  function
    | (-1, constant_coeff)::rest ->
	apron_linear_exprof_linear_expr_internal
	  (Some (coeff_of_z constant_coeff)) rest 
    | rest ->
	apron_linear_exprof_linear_expr_internal None rest;;

let apron_lincons0_of_linear_constraint
  ((expr, op) : (Formula.linear_expression *
	      [`GREATER_EQ|`GREATER|`LESS_EQ|`LESS |`EQUAL ])) : Lincons0.t =
  match op with
    | `GREATER_EQ -> Lincons0.make (apron_linexpr0_of_linear_expr expr) Lincons0.SUPEQ
    | `GREATER -> Lincons0.make (apron_linexpr0_of_linear_expr expr) Lincons0.SUP
    | `LESS_EQ -> Lincons0.make (apron_linexpr0_of_linear_expr (Formula.linear_expression_neg expr)) Lincons0.SUPEQ
    | `LESS -> Lincons0.make (apron_linexpr0_of_linear_expr (Formula.linear_expression_neg expr)) Lincons0.SUP

    | `EQUAL -> Lincons0.make (apron_linexpr0_of_linear_expr expr) Lincons0.EQ;;

let z_of_coeff = function
  | Coeff.Scalar (Scalar.Mpqf coeff) ->
  (if not (Z.equal (Q.get_den coeff) Z.one)
   then failwith "z_of_coeff: fractional"
   else Q.get_num coeff)
  | Coeff.Scalar (Scalar.Float x) -> failwith
      (Printf.sprintf "z_of_coeff: float <%g>" x)
  | Coeff.Interval _ -> failwith "z_of_coeff: interval";;

let apron_linexpr0_to_linear_expression (expr : Linexpr0.t) : Formula.linear_expression =
  let non_constant =
    (let o = ref [] in
      Linexpr0.iter
	(fun coeff id ->
	   let zv = z_of_coeff coeff in
	   if not (Z.is_zero zv)
	   then o := (id, zv)::!o) expr;
       List.rev !o)
  and constant = z_of_coeff (Linexpr0.get_cst expr) in
  let r =
    if Z.is_zero constant
    then non_constant
    else (-1, constant)::non_constant in
    assert(Formula.linear_expr_ok r);
    r;;

let apron_lincons0_to_linear_constraint (cons : Lincons0.t) :
  Formula.linear_expression * [`GREATER_EQ|`GREATER|`EQUAL ] =
  (apron_linexpr0_to_linear_expression cons.Lincons0.linexpr0),
  (match cons.Lincons0.typ with
     | Lincons0.SUP -> `GREATER
     | Lincons0.SUPEQ -> `GREATER_EQ
     | Lincons0.EQ -> `EQUAL
     | _ -> failwith "apron_lincons0_to_linear_constraint: bad constraint type out of Lincons0");;

let convert_minimize_in manager (l : Projection.minimize_in) =
  let dimension = Formula.linear_constraints_get_dimension l in
  let element = Abstract0.top manager 0 dimension in
    Abstract0.meet_lincons_array_with manager element
      (Array.map apron_lincons0_of_linear_constraint (Array.of_list l));
    element;;

let convert_minimize_out manager element =
  Abstract0.minimize manager element;
  Abstract0.canonicalize manager element;
  if Abstract0.is_bottom manager element
  then raise Projection.Empty_polyhedron
  else List.map apron_lincons0_to_linear_constraint
	  (Array.to_list (Abstract0.to_lincons_array manager element));;

let minimize_constraints manager (x : Projection.minimize_in):
    Projection.minimize_out=
  convert_minimize_out manager (convert_minimize_in manager x);;

let gen_range low next_high =
  Array.init (next_high - low) (fun i -> i+low);;

let project_minimize_constraints manager (x : Projection.minimize_in)
    (project_dim : int):
    Projection.minimize_out =
  let element = convert_minimize_in manager x in 
  let dim = (Abstract0.dimension manager element).Dim.reald in
    (if project_dim < dim
     then
       Abstract0.remove_dimensions_with manager element
	 { Dim.dim = gen_range project_dim dim;
	   Dim.intdim = 0;
	   Dim.realdim = dim - project_dim });
    convert_minimize_out manager element;;

let init () =
let manager = Polka.manager_alloc_strict() in
Multiple_choices.add_choice Projection.choices
  "apron-polka" 20
  { Projection. minimize_constraints =  minimize_constraints manager;
    Projection. project_minimize_constraints = project_minimize_constraints manager}
  "NewPolka library used through APRON";;
