open Formula

let rec linear_expr_to_ppl_nonconst (rest : PPL.linear_expr) :
    linear_expression -> PPL.linear_expr = function
      | (hv, hc)::tail ->
	  assert(hv >= 0);
	  linear_expr_to_ppl_nonconst
	    (PPL.linear_expr_add rest
	       (PPL.linear_expr_from_scaled_variable hc hv)) tail
      | [] -> rest;;

let linear_expr_to_ppl : linear_expression -> PPL.linear_expr = function
  | (-1, constant_coeff)::tail ->
      linear_expr_to_ppl_nonconst (PPL.linear_expr_from_z constant_coeff) tail
  | tail -> linear_expr_to_ppl_nonconst PPL.linear_expr_zero tail;;

let linear_constraint_to_ppl (l : linear_expression)
    (op : convex_comparison_op) : PPL.linear_constraint =
  PPL.constraint_make
    (linear_expr_to_ppl l)
    (match op with
       | `GREATER_EQ -> PPL.GREATER_EQ
       | `GREATER -> PPL.GREATER
       | `LESS_EQ -> PPL.LESS_EQ
       | `LESS -> PPL.LESS
       | `EQUAL -> PPL.EQUAL)
    PPL.linear_expr_zero;;

let linear_constraints_to_ppl :
    (linear_expression * convex_comparison_op) list ->
  PPL.linear_constraint list = List.map
  (fun (l, op) -> linear_constraint_to_ppl l op);;

let linear_constraint_from_ppl (max_dimension : int)
   (ct : PPL.linear_constraint) : 
    linear_expression * [`GREATER_EQ | `GREATER | `EQUAL ] =
  let (form, op, constant) = PPL.constraint_read ct in
    assert(linear_expr_ok form);
    List.iter (fun (d, _) -> assert (d < max_dimension)) form;

(*    List.iter (fun (k, v) ->
		 Printf.printf "caml k=%d v=%s\n" k (Z.to_string v);
	      ) form;
    Printf.printf "caml constant = %s\n" (Z.to_string constant);
    flush stdout;*)
    let l =
      if (Z.sgn constant) = 0
      then form
      else (-1, (Z.neg constant))::form in
      match op with
	| PPL.GREATER_EQ -> l, `GREATER_EQ
	| PPL.GREATER -> l, `GREATER
	| PPL.EQUAL -> l, `EQUAL
	| PPL.LESS | PPL.LESS_EQ -> failwith "bad constraint_kind";;

let linear_constraints_from_ppl (max_dimension : int) :
    PPL.linear_constraint list ->
  (linear_expression * [`GREATER_EQ | `GREATER | `EQUAL ]) list =
  List.map (linear_constraint_from_ppl max_dimension);;

let minimize_constraints
    (cts : (linear_expression *
	      [`GREATER_EQ|`GREATER|`LESS_EQ|`LESS |`EQUAL ]) list) :
    (linear_expression * [`GREATER_EQ|`GREATER|`EQUAL]) list =
  let dimension = linear_constraints_get_dimension cts in
  let poly =  PPL.nnc_poly_create dimension PPL.UNIVERSE in
    PPL.nnc_poly_add_constraints_and_minimize poly
      (linear_constraints_to_ppl cts);
    if PPL.nnc_poly_is_empty poly
    then raise Projection.Empty_polyhedron
    else
      linear_constraints_from_ppl dimension (PPL.nnc_poly_minimized_constraints poly);;

let project_minimize_constraints
    (cts : (linear_expression *
	      [`GREATER_EQ|`GREATER|`LESS_EQ|`LESS |`EQUAL ]) list)
    (projection_dimension : int) :
    (linear_expression * [`GREATER_EQ|`GREATER|`EQUAL]) list =
  let dimension = linear_constraints_get_dimension cts in
    assert(projection_dimension >= 0);
  let poly =  PPL.nnc_poly_create dimension PPL.UNIVERSE in
    PPL.nnc_poly_add_constraints poly
      (linear_constraints_to_ppl cts);
    (if projection_dimension < dimension
     then PPL.nnc_poly_remove_higher_space_dimensions
       poly projection_dimension);
    if PPL.nnc_poly_is_empty poly
    then raise Projection.Empty_polyhedron
    else
      linear_constraints_from_ppl projection_dimension (PPL.nnc_poly_minimized_constraints poly);;

Multiple_choices.add_choice Projection.choices
  "ppl" 10
  { Projection. minimize_constraints =  minimize_constraints;
    Projection. project_minimize_constraints = project_minimize_constraints }
  "PPL used directly";;
